import json
import os
from unittest.mock import MagicMock, patch
from django.contrib.auth.models import User
from django.test import TestCase
from datetime import date, timedelta

from esi.openapi_clients import ESIClientProvider
from django.core.cache import cache
from django.core.management import call_command
from django.utils import timezone
from esi.tests import NoSocketsTestCase
from httpx import RequestError, HTTPStatusError
from esi.exceptions import ESIErrorLimitException, HTTPClientError, HTTPNotModified, ESIBucketLimitException, HTTPServerError
from esi.rate_limiting import ESIRateLimits
from esi import app_settings
from esi import __title__, __url__, __version__
import httpx

from . import _generate_token, _store_as_Token
from .. import openapi_clients as oc

SPEC_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_openapi.json"
)
MODULE_PATH_PLUGINS = 'esi.aiopenapi3.plugins'


class TestClientFunctions(TestCase):
    def test_time_to_expiry_valid(self):
        expires = (
            timezone.now() + timedelta(seconds=120)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')
        ttl = oc._time_to_expiry(expires)

        # this shouldnt take more that 10 seconds
        self.assertGreater(ttl, 110)

    def test_time_to_expiry_invalid(self):
        # invalid format returns 0
        self.assertEqual(oc._time_to_expiry("not-a-date"), 0)

    def test_httpx_exceptions_valids(self):
        self.assertTrue(
            oc._httpx_exceptions(
                RequestError("Bad Request")
            )
        )

        response = MagicMock(status_code=502)
        exc = HTTPStatusError("msg", request=None, response=response)

        self.assertTrue(
            oc._httpx_exceptions(exc)
        )

        response.status_code = 400
        exc = HTTPStatusError("msg", request=None, response=response)

        self.assertFalse(
            oc._httpx_exceptions(exc)
        )

        self.assertFalse(
            oc._httpx_exceptions(
                ESIErrorLimitException(reset=10)
            )
        )

    def test_httpx_exceptions_invalid(self):
        self.assertFalse(
            oc._httpx_exceptions(
                "this is not an exception!"
            )
        )


class BuildUserAgentTests(TestCase):
    app_name = "TestApp"
    app_ver = "1.2.3"
    app_url = "https://tests.pass"

    def test_build_user_agent_with_url(self):
        ua = oc._build_user_agent(self.app_name, self.app_ver, self.app_url)

        expected_app_name = "TestApp"
        expected_title = "DjangoEsi"

        self.assertEqual(
            (
                f"{expected_app_name}/{self.app_ver} "
                f"({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{self.app_url})'} "
                f"{expected_title}/{__version__} (+{__url__})"
            ),
            ua
        )

    def test_enforce_pascal_case_for_ua_appname_with_space(self):
        """
        Test that the application name is converted to PascalCase in the User-Agent string when it contains spaces.

        :return:
        :rtype:
        """

        ua = oc._build_user_agent("test app", self.app_ver, self.app_url)

        expected_app_name = "TestApp"
        expected_title = "DjangoEsi"

        self.assertEqual(
            (
                f"{expected_app_name}/{self.app_ver} "
                f"({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{self.app_url})'} "
                f"{expected_title}/{__version__} (+{__url__})"
            ),
            ua
        )

    def test_enforce_pascal_case_for_ua_appname_with_hyphen(self):
        """
        Test that the application name is converted to PascalCase in the User-Agent string when it contains hyphens.

        :return:
        :rtype:
        """

        ua = oc._build_user_agent("test-app", self.app_ver, self.app_url)

        expected_app_name = "TestApp"
        expected_title = "DjangoEsi"

        self.assertEqual(
            (
                f"{expected_app_name}/{self.app_ver} "
                f"({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{self.app_url})'} "
                f"{expected_title}/{__version__} (+{__url__})"
            ),
            ua
        )

    def test_build_user_agent_without_url(self):
        ua = oc._build_user_agent(self.app_name, self.app_ver)

        expected_app_name = "TestApp"
        expected_title = "DjangoEsi"

        self.assertEqual(
            (
                f"{expected_app_name}/{self.app_ver} "
                f"({app_settings.ESI_USER_CONTACT_EMAIL}) "
                f"{expected_title}/{__version__} (+{__url__})"
            ),
            ua
        )


class BaseEsiOperationTests(TestCase):
    def setUp(self):
        self.page_param = MagicMock()
        self.page_param.name = "page"

        self.after_param = MagicMock()
        self.after_param.name = "after"

        self.before_param = MagicMock()
        self.before_param.name = "before"

        self.data_param = MagicMock()
        self.data_param.name = "data"

        self.lang_param = MagicMock()
        self.lang_param.name = "Accept-Language"

        self.body_param = MagicMock()
        self.body_param.name = "body"

        self.fake_op = MagicMock()
        self.fake_op.parameters = [
            self.data_param,
            self.lang_param
        ]
        self.fake_op.tags = ["test"]
        self.fake_op.operationId = "fake_op"
        self.fake_op.extensions = {}
        self.api = MagicMock(app_name="TestApp")
        self.op = oc.BaseEsiOperation(
            ("GET", "/fake_op", self.fake_op, {}),
            self.api
        )

    def test_non_unique_kwargs(self):
        op_1 = self.op(data="bar")
        key_1 = op_1._cache_key()
        op_2 = self.op(data="foo")
        key_2 = op_2._cache_key()
        self.assertNotEqual(key_1, key_2)

    def test_unique_kwargs(self):
        op_1 = self.op(data="foo")
        key_1 = op_1._cache_key()
        op_2 = self.op(data="foo")
        key_2 = op_2._cache_key()
        self.assertEqual(key_1, key_2)

    def test_extract_body(self):
        test_body = "something somethng something..."
        op = self.op(body=test_body)
        body = op._extract_body_param()
        self.assertEqual(test_body, body)

    def test_extract_body_exception(self):
        test_body = "something somethng something..."
        self.fake_op.requestBody = False
        op = self.op(body=test_body)
        with self.assertRaises(ValueError):
            op._extract_body_param()

    def test_extract_token(self):
        test_tkn = {"token": "token model goes here"}
        op = self.op(token=test_tkn)
        token = op._extract_token_param()
        self.assertEqual(test_tkn, token)

    def test_extract_token_exception_no_token_needed(self):
        self.op._kwargs = {"token": "token"}
        self.fake_op.security = None
        with self.assertRaises(ValueError):
            self.op._extract_token_param()

    def test_not_page_or_cursor_param(self):
        self.assertFalse(self.op._has_page_param())
        self.assertFalse(self.op._has_cursor_param())

    def test_has_page_param(self):
        self.fake_op.parameters += [self.page_param]
        op = self.op()
        self.assertTrue(op._has_page_param())

    def test_has_cursor_params(self):
        self.fake_op.parameters = [self.after_param]
        op = self.op()
        self.assertTrue(op._has_cursor_param())

        self.fake_op.parameters = [self.before_param]
        op = self.op()
        self.assertTrue(op._has_cursor_param())


class EsiOperationTests(TestCase):
    def setUp(self):
        self.op_mock = MagicMock()
        self.op_mock.parameters = []
        self.op_mock.tags = ["tag"]
        self.op_mock.operationId = "opid"
        self.op_mock.extensions = {}

        self.op_mock_rate = MagicMock()
        self.op_mock_rate.parameters = []
        self.op_mock_rate.tags = ["tag"]
        self.op_mock_rate.operationId = "opidRated"
        self.op_mock_rate.extensions = {
            "rate-limit": {
                "group": "test-group",
                "max-tokens": 100,
                "window-size": "5m"
            }
        }

        self.api_mock = MagicMock()
        self.api_mock.app_name = "TestApp"

        self.op = oc.EsiOperation(
            (
                "GET",
                "/url",
                self.op_mock,
                {}
            ),
            self.api_mock
        )

        self.op_rate_rated = oc.EsiOperation(
            (
                "GET",
                "/url",
                self.op_mock_rate,
                {}
            ),
            self.api_mock
        )

    @patch.object(oc.EsiOperation, "_make_request")
    def test_result_and_results(self, mock_make_request):
        data = {"data": "stuff"}
        mock_resp = MagicMock(status_code=200, headers={"Expires": "Wed, 1 July 2099 11:00:00 GMT"})
        mock_make_request.return_value = ({"Expires": "date"}, data, mock_resp)
        data_resp = self.op(foo="bar").result()
        self.assertEqual(data, data_resp)

    def test_esi_bucket_public(self):
        op = self.op_rate_rated(foo="bar")
        self.assertEqual(op.bucket.window, 300)
        self.assertEqual(op.bucket.slug, "test-group")
        self.assertEqual(op.bucket.limit, 100)

    def test_esi_bucket_limit(self):
        op = self.op_rate_rated(foo="bar")
        ESIRateLimits.set_bucket(op.bucket, 0)

        with self.assertRaises(ESIBucketLimitException):
            op.result()


class TestOpenapiClientProvider(NoSocketsTestCase):
    def setUp(self):
        self.app_name = "TestsApp"
        self.app_ver = "1.2.3"
        self.app_url = "https://tests.pass"
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Status"],
            spec_file=SPEC_PATH
        )
        cache.clear()

    def test_str(self):
        self.assertIn(self.esi._ua_appname, str(self.esi))
        self.assertIn(self.esi._ua_version, str(self.esi))

    def test_compatibilitydate_date_to_string(self):
        testdate_1 = date(2024, 1, 1)
        testdate_2 = date(2025, 8, 26)

        self.assertEqual("2024-01-01", ESIClientProvider._date_to_string(testdate_1))
        self.assertEqual("2025-08-26", ESIClientProvider._date_to_string(testdate_2))

    def test_compatibility_date_as_date(self):
        testdate = date(2024, 1, 1)
        app_name = "TestsApp"
        app_ver = "1.2.3"
        app_url = "https://tests.pass"

        esi = ESIClientProvider(
            ua_appname=app_name,
            ua_url=app_url,
            ua_version=app_ver,
            compatibility_date=testdate,
            spec_file=SPEC_PATH
        )
        self.assertEqual(esi._compatibility_date, ESIClientProvider._date_to_string(testdate))


    @patch.object(httpx.Client, "send")
    def test_ua(self, send: MagicMock):
        send.return_value = httpx.Response(
            200,
            json={
                "players": 1234,
                "server_version": "1234",
                "start_time": "2029-09-19T11:02:08Z"
            },
            request=httpx.Request("GET", "test"),
        )

        status = self.esi.client.Status.GetStatus().result()
        call_args, call_kwargs = send.call_args

        expected_app_name = "TestsApp"
        expected_title = 'DjangoEsi'

        self.assertEqual(
            call_args[0].headers["user-agent"],
            (
                f"{expected_app_name}/{self.app_ver} "
                f"({app_settings.ESI_USER_CONTACT_EMAIL}{f'; +{self.app_url})'} "
                f"{expected_title}/{__version__} (+{__url__})"
            )
        )
        self.assertEqual(status.players, 1234)

    @patch(MODULE_PATH_PLUGINS + '.settings.DEBUG', True)
    def test_no_tag_no_op_debug(self):
        app_name = "TestsApp"
        app_ver = "1.2.3"
        app_url = "https://tests.pass"

        esi = ESIClientProvider(
            ua_appname=app_name,
            ua_url=app_url,
            ua_version=app_ver,
            compatibility_date="2020-01-01",
            spec_file=SPEC_PATH
        )
        self.assertIsNotNone(esi.client.Status)
        self.assertIsNotNone(esi.client.Status.GetStatus)

    @patch(MODULE_PATH_PLUGINS + '.settings.DEBUG', True)
    def test_tag_no_op_debug(self):
        app_name = "TestsApp"
        app_ver = "1.2.3"
        app_url = "https://tests.pass"

        esi = ESIClientProvider(
            ua_appname=app_name,
            ua_url=app_url,
            ua_version=app_ver,
            compatibility_date="2020-01-01",
            tags=["Status"],
            spec_file=SPEC_PATH
        )
        self.assertIsNotNone(esi.client.Status)
        self.assertIsNotNone(esi.client.Status.GetStatus)

    @patch(MODULE_PATH_PLUGINS + '.settings.DEBUG', True)
    def test_no_tag_op_debug(self):
        app_name = "TestsApp"
        app_ver = "1.2.3"
        app_url = "https://tests.pass"

        esi = ESIClientProvider(
            ua_appname=app_name,
            ua_url=app_url,
            ua_version=app_ver,
            compatibility_date="2020-01-01",
            operations=["GetStatus"],
            spec_file=SPEC_PATH
        )
        self.assertIsNotNone(esi.client.Status)
        self.assertIsNotNone(esi.client.Status.GetStatus)

    @patch(MODULE_PATH_PLUGINS + '.settings.DEBUG', False)
    def test_no_tag_no_op_no_debug(self):
        app_name = "TestsApp"
        app_ver = "1.2.3"
        app_url = "https://tests.pass"

        with self.assertRaises(AttributeError):
            esi = ESIClientProvider(
                ua_appname=app_name,
                ua_url=app_url,
                ua_version=app_ver,
                compatibility_date="2020-01-01",
                spec_file=SPEC_PATH
            )
            esi.client

    @patch.object(httpx.Client, "send")
    def test_no_bucket(self, send: MagicMock):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Universe"],
            spec_file=SPEC_PATH
        )

        send.return_value = httpx.Response(
            200,
            json=[1, 2, 3, 4],
            request=httpx.Request("GET", "test"),
        )

        types = self.esi.client.Universe.GetUniverseTypes().result()
        self.assertEqual(len(types), 4)

    @patch.object(httpx.Client, "send")
    def test_etag_hit_cached(self, send: MagicMock):
        etag = "'123456789abcdef123456789abcdef'"

        expires = (
            timezone.now() + timedelta(minutes=5)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')

        send.return_value = httpx.Response(
            200,
            json={
                "players": 1234,
                "server_version": "1234",
                "start_time": "2029-09-19T11:02:08Z"
            },
            headers={
                "etag": etag,
                "expires": expires
            },
            request=httpx.Request(
                "GET",
                "test",
            ),
        )

        self.esi.client.Status.GetStatus().result()

        with self.assertRaises(HTTPNotModified):
            self.esi.client.Status.GetStatus().result()

    @patch.object(httpx.Client, "send")
    def test_pages_etag_partial_hit_cached(self, send: MagicMock):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Assets"],
            spec_file=SPEC_PATH
        )

        etag = "'123456789abcdef123456789abcdef'"

        expires = (
            timezone.now() + timedelta(minutes=5)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')

        send.return_value = httpx.Response(
            200,
            json=[
                {
                    "is_blueprint_copy": None,
                    "is_singleton": True,
                    "item_id": 1234,
                    "location_flag": "HiSlot6",
                    "location_id": 123456,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": 123456
                }
            ],
            headers={
                "etag": etag,
                "expires": expires,
                "x-pages": "5"
            },
            request=httpx.Request(
                "GET",
                "test",
            ),
        )

        # get one page only to have a single etag filled in
        hit_1 = self.esi.client.Assets.GetCharactersCharacterIdAssets(
            character_id=123,
            token="FakeToken",
            page=1
        ).result()

        # now i want all of them and give them to me
        assets = self.esi.client.Assets.GetCharactersCharacterIdAssets(
            character_id=123,
            token="FakeToken"
        ).results()

        print(vars(send))

        self.assertEqual(len(assets), 5)
        self.assertEqual(send.call_count, 5) # only hit esi for 5 pages, use cache for rest.

        with self.assertRaises(HTTPNotModified):
            # now we have all the data and throw this error
            assets = self.esi.client.Assets.GetCharactersCharacterIdAssets(
                character_id=123,
                token="FakeToken"
            ).results()

    @patch.object(httpx.Client, "send")
    def test_etag_not_hit_cached(self, send: MagicMock):
        etag = "'123456789abcdef123456789abcdef'"

        expires = (
            timezone.now() + timedelta(minutes=5)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')

        send.return_value = httpx.Response(
            200,
            json={
                "players": 1234,
                "server_version": "1234",
                "start_time": "2029-09-19T11:02:08Z"
            },
            headers={
                "etag": etag,
                "expires": expires
            },
            request=httpx.Request(
                "GET",
                "test",
            ),
        )

        self.esi.client.Status.GetStatus().result()

        result = self.esi.client.Status.GetStatus().result(use_etag=False)
        self.assertEqual(result.players, 1234)

    @patch.object(httpx.Client, "send")
    def test_force_refresh(self, send: MagicMock):
        etag = "'123456789abcdef123456789abcdef'"

        expires = (
            timezone.now() + timedelta(minutes=5)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')

        send.return_value = httpx.Response(
            200,
            json={
                "players": 1234,
                "server_version": "1234",
                "start_time": "2029-09-19T11:02:08Z"
            },
            headers={
                "etag": etag,
                "expires": expires
            },
            request=httpx.Request(
                "GET",
                "test",
            ),
        )

        self.esi.client.Status.GetStatus().result()

        result = self.esi.client.Status.GetStatus().result(force_refresh=True)
        self.assertEqual(result.players, 1234)
        self.assertEqual(send.call_count, 2)

    @patch.object(httpx.Client, "send")
    def test_404(self, send: MagicMock):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Universe"],
            spec_file=SPEC_PATH
        )

        send.return_value = httpx.Response(
            404,
            json={
                "error": "error"
            },
            headers={
                "X-RateLimit-Reset": "15",
                "X-RateLimit-Remaining": "0"
            },
            request=httpx.Request(
                "GET",
                "/universe/types"
            ),
        )

        with self.assertRaises(HTTPClientError):
            self.esi.client.Universe.GetUniverseTypes().result()

    @patch.object(httpx.Client, "send")
    def test_420(self, send: MagicMock):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Universe"],
            spec_file=SPEC_PATH
        )

        send.return_value = httpx.Response(
            420,
            json={
                "error": "error"
            },
            headers={
                "X-RateLimit-Reset": "15",
                "X-RateLimit-Remaining": "0"
            },
            request=httpx.Request(
                "GET",
                "/universe/types"
            ),
        )

        with self.assertRaises(ESIErrorLimitException):
            self.esi.client.Universe.GetUniverseTypes().result()

        self.assertGreater(cache.get("esi_error_limit_reset"), 10)

    @patch.object(httpx.Client, "send")
    def test_420_past(self, send: MagicMock):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Universe"],
            spec_file=SPEC_PATH
        )

        send.return_value = httpx.Response(
            420,
            json={
                "error": "error"
            },
            headers={
                "X-RateLimit-Remaining": "0"
            },
            request=httpx.Request(
                "GET",
                "/universe/types"
            ),
        )

        with self.assertRaises(ESIErrorLimitException):
            self.esi.client.Universe.GetUniverseTypes().result()

        self.assertIsNone(cache.get("esi_error_limit_reset"))

    @patch.object(httpx.Client, "send")
    def test_rate_bucket(self, send: MagicMock):
        send.return_value = httpx.Response(
            200,
            json={
                "players": 1234,
                "server_version": "1234",
                "start_time": "2029-09-19T11:02:08Z"
            },
            headers={
                "x-ratelimit-group": "status",
                "x-ratelimit-used": "2",
                "x-ratelimit-remaining": "598",
                "x-ratelimit-limit": "600/15m",
            },
            request=httpx.Request(
                "GET",
                "/status"
            ),
        )
        self.esi.client.Status.GetStatus().result()
        self.assertEqual(
            ESIRateLimits.get_bucket(self.esi.client.Status.GetStatus().bucket),
            598
        )

    @patch.object(httpx.Client, "send")
    def test_server_error(self, send: MagicMock):
        send.return_value = httpx.Response(
            520,
            json={
                "error": "error"
            },
            headers={
                "x-ratelimit-group": "status",
                "x-ratelimit-used": "5",
                "x-ratelimit-remaining": "595",
                "x-ratelimit-limit": "600/15m",
            },
            request=httpx.Request(
                "GET",
                "/status"
            ),
        )
        with self.assertRaises(HTTPServerError):
            self.esi.client.Status.GetStatus().result()

    def test_minified_op_not_found(self):
        with self.assertRaises(AttributeError):
            self.esi.client.Universe.GetUniverseTypes()

    def test_minified_op_not_found(self):
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Universe"],
            spec_file=SPEC_PATH
        )

        with self.assertRaises(AttributeError):
            self.esi.client.Universe.GetUniverseAncestries()

    def test_rate_bucket_found_in_spec(self):

        op = self.esi.client.Status.GetStatus()

        self.assertIsNotNone(op.bucket)
        self.assertEqual(op.bucket.limit, 600)
        self.assertEqual(op.bucket.slug, "status")
        self.assertEqual(op.bucket.window, 900)

    def test_rate_bucket_hit(self):
        op = self.esi.client.Status.GetStatus()

        ESIRateLimits.set_bucket(op.bucket, 0)

        with self.assertRaises(ESIBucketLimitException):
            op.result()

    def test_global_limit_hit(self):
        op = self.esi.client.Status.GetStatus()

        cache.set("esi_error_limit_reset", 15, 15)

        with self.assertRaises(ESIErrorLimitException):
            op.result()

    @patch.object(httpx.Client, "send")
    def test_load_sync(self, send: MagicMock):
        esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Status"]
        )
        cache.clear()
        expires = (
            timezone.now() + timedelta(minutes=5)
        ).strftime('%a, %d %b %Y %H:%M:%S %Z')

        spec = None
        with open(SPEC_PATH) as f:
            spec = json.load(f)

        send.return_value = httpx.Response(
            200,
            json=spec,
            headers={
                "expires": expires
            },
            request=httpx.Request(
                "GET",
                "test",
            ),
        )

        esi.client
        self.assertIsNotNone(esi.client.Status)


class TestTokenisedEndpoints(TestCase):
    def setUp(self):
        self.app_name = "TestsApp"
        self.app_ver = "1.2.3"
        self.app_url = "https://tests.pass"
        self.esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Character", "Assets"],
            spec_file=SPEC_PATH
        )
        cache.clear()

        self.character_id = 1000
        character_name = 'Bruce Wayne'

        self.user = User.objects.create_user(
            character_name,
            'abc@example.com',
            'password'
        )

        self.token_str = _store_as_Token(
            _generate_token(
                character_id=self.character_id,
                character_name=character_name,
                scopes=['esi-assets.read_assets.v1']
            ),
            self.user
        )

        self.token_note = _store_as_Token(
            _generate_token(
                character_id=self.character_id,
                character_name=character_name,
                scopes=['esi-characters.read_notifications.v1']
            ),
            self.user
        )

        self.resp_note = httpx.Response(
            200,
            json=[{
                "is_read": None,
                "notification_id": 123456,
                "sender_id": 12345,
                "sender_type": "corporation",
                "text": "text...",
                "timestamp": "2024-11-25T05:37:00Z",
                "type": "CorpAllBillMsg"
            },
                {
                "is_read": None,
                "notification_id": 123457,
                "sender_id": 123456,
                "sender_type": "corporation",
                "text": "text...",
                "timestamp": "2024-11-25T01:43:00Z",
                "type": "MoonminingExtractionStarted"
            },
            ],
            headers={
                "x-ratelimit-group": "char-notification",
                "x-ratelimit-used": "5",
                "x-ratelimit-remaining": "595",
                "x-ratelimit-limit": "600/15m",
            },
            request=httpx.Request(
                "GET",
                "/characters/{character_id}/notifications"
            ),
        )

        self.resp_asset = httpx.Response(
            200,
            json=[
                {
                    "is_blueprint_copy": None,
                    "is_singleton": False,
                    "item_id": 12346,
                    "location_flag": "Hangar",
                    "location_id": 60003760,
                    "location_type": "station",
                    "quantity": 999,
                    "type_id": 3764
                },
                {
                    "is_blueprint_copy": None,
                    "is_singleton": False,
                    "item_id": 12345,
                    "location_flag": "Hangar",
                    "location_id": 60003760 ,
                    "location_type": "station",
                    "quantity": 999,
                    "type_id": 11567
                }
            ],
            headers={
                "x-esi-error-limit-remain": "100",
                "x-esi-error-limit-reset": "60",
                "x-pages": "5"
            },
            request=httpx.Request(
                "GET",
                "/characters/{character_id}/notifications"
            ),
        )

    @patch.object(httpx.Client, "send")
    def test_no_token_provided(self, send: MagicMock):
        send.return_value = self.resp_note
        with self.assertRaises(ValueError):
            self.esi.client.Character.GetCharactersCharacterIdNotifications(
                character_id=self.character_id
            ).result()
        # Didn't hit ESI
        self.assertEqual(send.call_count, 0)

    @patch.object(httpx.Client, "send")
    def test_good_token_provided(self, send: MagicMock):
        send.return_value = self.resp_note
        results = self.esi.client.Character.GetCharactersCharacterIdNotifications(
            character_id=self.character_id,
            token=self.token_note
        ).result()
        self.assertEqual(send.call_count, 1)
        self.assertEqual(len(results), 2)

    @patch.object(httpx.Client, "send")
    def test_bad_token_provided(self, send: MagicMock):
        send.return_value = self.resp_note
        with self.assertRaises(ValueError):
            self.esi.client.Character.GetCharactersCharacterIdNotifications(
                character_id=self.character_id,
                token=self.token_str
            ).result()
        # Didn't hit ESI
        self.assertEqual(send.call_count, 0)

    @patch.object(httpx.Client, "send")
    def test_token_pages(self, send: MagicMock):
        send.return_value = self.resp_asset
        results = self.esi.client.Assets.GetCharactersCharacterIdAssets(
            character_id=self.character_id,
            token=self.token_str
        ).results()
        self.assertEqual(send.call_count, 5)
        self.assertEqual(len(results), 10)

class TestTokenisedEndpoints(TestCase):
    def setUp(self):
        cache.clear()
        self.app_name = "TestsApp"
        self.app_ver = "1.2.3"
        self.app_url = "https://tests.pass"
        spec = {}
        with open(SPEC_PATH) as f:
            spec = json.load(f)

        self.resp = httpx.Response(
            200,
            json=spec,
            headers={
                "x-esi-error-limit-remain": "100",
                "x-esi-error-limit-reset": "60"
            },
            request=httpx.Request(
                "GET",
                "https://esi.evetech.net/meta/openapi.json"
            ),
        )

    @patch.object(httpx.Client, "get")
    def test_load_spec(self, get: MagicMock):
        get.return_value = self.resp

        esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Character", "Assets"],
        )
        esi.client
        self.assertEqual(get.call_count, 1)

        esi2 = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-10-10",
            tags=["status"],
        )
        esi2.client
        self.assertEqual(get.call_count, 2)

        esi3 = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Character"],
        )
        esi3.client
        self.assertEqual(get.call_count, 2)

        esi4 = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-10-10",
            tags=["Assets"],
        )
        esi4.client
        self.assertEqual(get.call_count, 2)

    @patch.object(httpx.Client, "get")
    def test_purge_cache_load_spec(self, get: MagicMock):
        get.return_value = self.resp

        esi = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Character", "Assets"],
        )
        esi.client
        self.assertEqual(get.call_count, 1)

        call_command("esi_clear_spec_cache")

        esi2 = ESIClientProvider(
            ua_appname=self.app_name,
            ua_url=self.app_url,
            ua_version=self.app_ver,
            compatibility_date="2020-01-01",
            tags=["Character", "Assets"],
        )
        esi2.client
        self.assertEqual(get.call_count, 2)



# Generated by Django 3.2.12 on 2022-05-16 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0011_add_token_indices'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='token_type',
            field=models.CharField(choices=[('character', 'Character'), ('corporation', 'Corporation')], default='character', help_text='The applicable range of the token.', max_length=100),
        ),
    ]



import datetime as dt
from hashlib import md5

from httpx import URL, Client, Response
from httpx._client import USE_CLIENT_DEFAULT, UseClientDefault
from httpx._types import (
    AuthTypes, CookieTypes, HeaderTypes, QueryParamTypes, RequestExtensions,
    TimeoutTypes,
)

from django.core.cache import cache
from django.utils import timezone


class SpecCachingClient(Client):

    @staticmethod
    def _is_spec(url: URL | str) -> bool:
        return "https://esi.evetech.net/meta/openapi" in str(url)

    @staticmethod
    def _get_seconds_to_dt() -> int:
        """
        Get the number of seconds until 11:30 AM the next day.

        :return:
        :rtype:
        """

        expire_time = timezone.now()
        target = expire_time.replace(hour=11, minute=30, second=0, microsecond=0)

        if expire_time >= target:
            target += dt.timedelta(days=1)

        return int((target - expire_time).total_seconds())

    def _get_spec_cache(self, url, ):
        return cache.get(self._get_api_cache_key(url), False)

    def _set_spec_cache(self, url, body):
        ttl = self._get_seconds_to_dt()
        cache.set(self._get_api_cache_key(url), body, ttl)

    def _get_api_cache_key(self, url: str) -> str:
        """
            provide a unique token for the spec url with compat date.
        """
        compat_date = self.headers.get("X-Compatibility-Date", None)
        return f"ESI_API_CACHE_{md5(f'{url}-{compat_date}'.encode()).hexdigest()}"

    def get(
        self,
        url: URL | str,
        *,
        params: QueryParamTypes | None = None,
        headers: HeaderTypes | None = None,
        cookies: CookieTypes | None = None,
        auth: AuthTypes | UseClientDefault | None = USE_CLIENT_DEFAULT,
        follow_redirects: bool | UseClientDefault = USE_CLIENT_DEFAULT,
        timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT,
        extensions: RequestExtensions | None = None,
    ) -> Response:
        if self._is_spec(url):
            _spec = self._get_spec_cache(url)
            if not _spec:
                _spec = super().get(
                    url,
                    params=params,
                    headers=headers,
                    cookies=cookies,
                    auth=auth,
                    follow_redirects=follow_redirects,
                    timeout=timeout,
                    extensions=extensions
                )
                self._set_spec_cache(url, _spec)
            return _spec
        return super().get(
            url,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
            extensions=extensions
        )



from bravado.exception import (
    HTTPBadGateway,
    HTTPBadRequest,
    HTTPForbidden,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPNotFound,
    HTTPServiceUnavailable,
    HTTPUnauthorized,
)


def create_http_error(http_code: int, text: str = None) -> Exception:
    """Build a HTTP exception for django-esi from given http code."""
    exc_map = {
        400: HTTPBadRequest,
        401: HTTPUnauthorized,
        403: HTTPForbidden,
        404: HTTPNotFound,
        500: HTTPInternalServerError,
        502: HTTPBadGateway,
        503: HTTPServiceUnavailable,
        504: HTTPGatewayTimeout,
    }
    try:
        http_exc = exc_map[http_code]
    except KeyError:
        raise NotImplementedError(f"Unknown http code: {http_code}") from None
    if not text:
        text = "Test exception"
    return http_exc(response=BravadoResponseStub(http_code, text))


class BravadoResponseStub:
    """Stub for IncomingResponse in bravado, e.g. for HTTPError exceptions."""

    def __init__(
        self, status_code, reason="", text="", headers=None, raw_bytes=None
    ) -> None:
        self.status_code = status_code
        self.reason = reason
        self.text = text
        self.headers = headers if headers else dict()
        self.raw_bytes = raw_bytes

    def __str__(self) -> str:
        return f"{self.status_code} {self.reason}"



import dataclasses

from aiopenapi3.errors import HTTPServerError as base_HTTPServerError
from aiopenapi3.errors import HTTPClientError as base_HTTPClientError
from aiopenapi3.errors import HTTPError


class ESIErrorLimitException(Exception):
    """ESI Global Error Limit Exceeded
    https://developers.eveonline.com/docs/services/esi/best-practices/#error-limit
    """

    def __init__(self, reset=None, *args, **kwargs) -> None:
        self.reset = reset
        msg = kwargs.get("message") or (
            f"ESI Error limited. Reset in {reset} seconds." if reset else "ESI Error limited."
        )
        super().__init__(msg, *args)


class ESIBucketLimitException(Exception):
    """Endpoint (Bucket) Specific Rate Limit Exceeded"""

    def __init__(self, bucket, reset: float = 0, *args, **kwargs) -> None:
        self.bucket = bucket
        self.reset = reset if reset else self.bucket.window
        msg = kwargs.get("message") or f"ESI bucket limit reached for {bucket}."
        super().__init__(msg, *args)


@dataclasses.dataclass(repr=False)
class HTTPNotModified(HTTPError):
    """The HTTP Status is 304"""

    status_code: int
    headers: dict[str, str]

    def __str__(self):
        return f"""<{self.__class__.__name__} {self.status_code} {self.headers}>"""


@dataclasses.dataclass(repr=False)
class HTTPClientError(base_HTTPClientError):
    """HTTP Response Code 4xx"""
    pass


@dataclasses.dataclass(repr=False)
class HTTPServerError(base_HTTPServerError):
    """HTTP Response Code 5xx"""
    pass



"""unit tests for esi"""

from datetime import timedelta
from unittest.mock import patch, Mock

from django.core.exceptions import ImproperlyConfigured
from django.contrib.auth.models import User
from django.utils import timezone
from django.test import TestCase

from oauthlib.oauth2.rfc6749.errors import InvalidGrantError, \
    MissingTokenError, InvalidClientError, InvalidTokenError, \
    InvalidClientIdError

from esi.errors import TokenInvalidError, NotRefreshableTokenError, \
    TokenExpiredError, IncompleteResponseError, TokenError
from esi.models import Token

from . import _generate_token, _store_as_Token
from .factories_2 import CallbackRedirectFactory, ScopeFactory, TokenFactory


MODULE_PATH = "esi.models"


class TestScope(TestCase):
    def test_should_return_str(self):
        # given
        obj = ScopeFactory.build(name="dummy_scope")
        # when/then
        self.assertEqual(str(obj), "dummy_scope")

    def test_should_return_friendly_name_1(self):
        # given
        obj = ScopeFactory.build(name="dummy_scope")
        # when/then
        self.assertEqual(obj.friendly_name, "dummy_scope")

    def test_should_return_friendly_name_2(self):
        # given
        obj = ScopeFactory.build(name="test.dummy_scope.test")
        # when/then
        self.assertEqual(obj.friendly_name, "dummy scope")


class TestToken2(TestCase):
    def test_should_return_str(self):
        # given
        scope = ScopeFactory(name="dummy_scope")
        token = TokenFactory(character_name="Bruce Wayne", scopes=[scope])
        # when/then
        self.assertEqual(str(token), "Bruce Wayne - dummy_scope")

    def test_should_return_str_not_created(self):
        # given
        token = TokenFactory.build(character_name="Bruce Wayne")
        # when/then
        self.assertEqual(str(token), "Bruce Wayne - ")

    def test_should_return_repr(self):
        # given
        scope = ScopeFactory(name="dummy_scope")
        token = TokenFactory(
            id=99, character_id=1001, character_name="Bruce Wayne", scopes=[scope]
        )
        # when/then
        self.assertEqual(repr(token), "<Token(id=99): 1001, Bruce Wayne>")

    def test_should_find_token(self):
        # given
        scope = ScopeFactory(name="dummy_scope")
        token = TokenFactory(character_id=1001, scopes=[scope])
        TokenFactory()
        # when
        result = Token.get_token(1001, ['dummy_scope'])
        # then
        self.assertEqual(token, result)

    def test_should_return_false_when_token_not_found(self):
        # given
        scope = ScopeFactory(name="dummy_scope")
        TokenFactory(character_id=1001, scopes=[scope])
        # when
        result = Token.get_token(1001, ['unknown'])
        # then
        self.assertFalse(result)

    def test_should_be_able_to_refresh(self):
        # given
        token = TokenFactory.build()
        # when/then
        self.assertTrue(token.can_refresh)

    def test_should_not_be_able_to_refresh(self):
        # given
        token = TokenFactory.build(refresh_token=None)
        # when/then
        self.assertFalse(token.can_refresh)


@patch(MODULE_PATH + ".Token.refresh", spec=True)
class TestTokenRefreshOrDelete(TestCase):
    def test_should_refresh_token(self, mock_token_refresh):
        # given
        token = TokenFactory()
        # when
        token.refresh_or_delete()
        # then
        self.assertTrue(mock_token_refresh.called)
        self.assertTrue(Token.objects.filter(pk=token.pk).exists())

    def test_should_delete_token_with_errors(self, mock_token_refresh):
        # given
        mock_token_refresh.side_effect = TokenError
        token = TokenFactory()
        # when
        token.refresh_or_delete()
        # then
        self.assertTrue(mock_token_refresh.called)
        self.assertFalse(Token.objects.filter(pk=token.pk).exists())

    def test_should_not_delete_token_with_incomplete_response_error(self, mock_token_refresh):
        # given
        mock_token_refresh.side_effect = IncompleteResponseError
        token = TokenFactory()
        # when
        with self.assertRaises(IncompleteResponseError):
            token.refresh_or_delete()
        # then
        self.assertTrue(mock_token_refresh.called)
        self.assertTrue(Token.objects.filter(pk=token.pk).exists())


class TestToken(TestCase):
    def setUp(self):

        character_id = 1000
        character_name = 'Bruce Wayne'

        self.user = User.objects.create_user(
            character_name,
            'abc@example.com',
            'password'
        )
        self.token = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=['esi-universe.read_structures.v1']
            ),
            self.user
        )

    @patch(MODULE_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    def test_expires(self):
        self.assertEqual(
            self.token.created + timedelta(seconds=120),
            self.token.expires
        )

    @patch(MODULE_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    def test_not_expired(self):
        self.assertFalse(self.token.expired)

    @patch(MODULE_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    def test_has_expired(self):
        self.token.created -= timedelta(121)
        self.assertTrue(self.token.expired)

    def test_refresh_normal_1(self):
        mock_auth = Mock()
        mock_session = Mock()
        mock_session.refresh_token.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }

        self.token.refresh(mock_session, mock_auth)
        self.assertEqual(
            self.token.refresh_token, 'refresh_token_2'
        )
        self.assertEqual(
            self.token.access_token, 'access_token_2'
        )
        self.assertGreaterEqual(
            self.token.created, timezone.now() - timedelta(seconds=60)
        )

    @patch(MODULE_PATH + '.HTTPBasicAuth', autospec=True)
    @patch(MODULE_PATH + '.OAuth2Session', autospec=True)
    def test_refresh_normal_2(self, mock_OAuth2Session, mock_HTTPBasicAuth):
        mock_session = Mock()
        mock_session.refresh_token.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        mock_OAuth2Session.return_value = mock_session

        self.token.refresh()
        self.assertEqual(
            self.token.refresh_token,
            'refresh_token_2'
        )
        self.assertEqual(
            self.token.access_token,
            'access_token_2'
        )
        self.assertGreaterEqual(
            self.token.created,
            timezone.now() - timedelta(seconds=60))

    def test_valid_access_token(self):
        self.assertFalse(self.token.expired)
        self.assertEqual(self.token.valid_access_token(), 'access_token')

    @patch(MODULE_PATH + '.HTTPBasicAuth', autospec=True)
    @patch(MODULE_PATH + '.OAuth2Session', autospec=True)
    @patch(MODULE_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    def test_valid_access_token_refresh(self, mock_OAuth2Session, mock_HTTPBasicAuth):
        mock_session = Mock()
        mock_session.refresh_token.return_value = {
            'access_token': 'access_token_new',
            'refresh_token': 'refresh_token_2'
        }
        mock_OAuth2Session.return_value = mock_session

        self.token.created -= timedelta(121)
        self.assertTrue(self.token.expired)
        self.assertEqual(
            self.token.valid_access_token(), 'access_token_new')

    @patch(MODULE_PATH + '.HTTPBasicAuth', autospec=True)
    @patch(MODULE_PATH + '.OAuth2Session', autospec=True)
    @patch(MODULE_PATH + '.app_settings.ESI_TOKEN_VALID_DURATION', 120)
    def test_valid_access_token_cant_refresh(
        self, mock_OAuth2Session, mock_HTTPBasicAuth
    ):
        self.token.refresh_token = None
        self.token.created -= timedelta(121)
        self.assertTrue(self.token.expired)
        with self.assertRaises(TokenExpiredError):
            self.token.valid_access_token()

    def test_refresh_errors_1(self):
        mock_auth = Mock()
        mock_session = Mock()
        mock_session.refresh_token.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        self.token.refresh_token = None
        with self.assertRaises(NotRefreshableTokenError):
            self.token.refresh(mock_session, mock_auth)

    def test_refresh_errors_2(self):
        mock_auth = Mock()
        mock_session = Mock()

        mock_session.refresh_token.side_effect = InvalidGrantError
        with self.assertRaises(TokenInvalidError):
            self.token.refresh(mock_session, mock_auth)

        mock_session.refresh_token.side_effect = InvalidTokenError
        with self.assertRaises(TokenInvalidError):
            self.token.refresh(mock_session, mock_auth)

        mock_session.refresh_token.side_effect = InvalidClientIdError
        with self.assertRaises(TokenInvalidError):
            self.token.refresh(mock_session, mock_auth)

        mock_session.refresh_token.side_effect = MissingTokenError
        with self.assertRaises(IncompleteResponseError):
            self.token.refresh(mock_session, mock_auth)

        mock_session.refresh_token.side_effect = InvalidClientError
        with self.assertRaises(ImproperlyConfigured):
            self.token.refresh(mock_session, mock_auth)

    @patch(MODULE_PATH + '.esi_client_factory', autospec=True)
    def test_get_esi_client(self, mock_esi_client):
        mock_esi_client.return_value = "Johnny"
        x = self.token.get_esi_client()
        self.assertEqual(x, "Johnny")
        self.assertEqual(mock_esi_client.call_count, 1)

    """
    @patch('esi.managers.TokenManager')
    def test_get_token_data(self, mock_decode_jwt):
        mock_decode_jwt._decode_jwt.return_value = \
            _generate_token(
                99, 'Bruce Wayne', scopes=[
                    'esi-calendar.read_calendar_events.v1',
                    'esi-location.read_location.v1',
                    'esi-location.read_ship_type.v1',
                    'esi-unknown-scope'
                ]
            )
        data = self.token.get_token_data(access_token='access_token_2')
        self.assertEqual(data['name'], "Bruce Wayne")
    """
    @patch(MODULE_PATH + '.Token.get_token_data')
    def test_update_token_data_normal_1(self, mock_get_token_data):
        mock_get_token_data.return_value = _generate_token(99, 'Bruce Wayne')
        self.token.update_token_data()
        self.token.refresh_from_db()
        self.assertEqual(
            self.token.character_id,
            99
        )
        self.assertEqual(
            self.token.character_name,
            'Bruce Wayne'
        )
        self.assertEqual(
            self.token.character_owner_hash,
            'character_owner_hash'
        )
        self.assertEqual(
            self.token.token_type,
            'character'
        )

    @patch(MODULE_PATH + '.HTTPBasicAuth', autospec=True)
    @patch(MODULE_PATH + '.OAuth2Session', autospec=True)
    @patch(MODULE_PATH + '.Token.get_token_data')
    def test_update_token_data_normal_2(
        self,
        mock_get_token_data,
        mock_OAuth2Session,
        mock_HTTPBasicAuth
    ):
        mock_session = Mock()
        mock_session.refresh_token.return_value = {
            'access_token': 'access_token_2',
            'refresh_token': 'refresh_token_2'
        }
        mock_OAuth2Session.return_value = mock_session

        mock_get_token_data.return_value = {
            'character_id': 99,
            'name': 'CharacterName',
            'owner': 'CharacterOwnerHash',
            'token_type': 'character',
        }
        self.token.created -= timedelta(121)

        self.token.update_token_data()
        self.token.refresh_from_db()
        self.assertEqual(
            self.token.character_id,
            99
        )

    @patch(MODULE_PATH + '.Token.get_token_data')
    def test_update_token_data_normal_3(self, mock_get_token_data):
        mock_get_token_data.return_value = {
            'character_id': 99,
            'name': 'CharacterName',
            'owner': 'CharacterOwnerHash',
            'token_type': 'Character',
        }
        self.token.update_token_data(commit=False)
        self.assertEqual(
            self.token.character_id,
            99
        )
        self.assertEqual(
            self.token.character_name,
            'CharacterName'
        )
        self.assertEqual(
            self.token.character_owner_hash,
            'CharacterOwnerHash'
        )
        self.assertEqual(
            self.token.token_type,
            'Character'
        )

    @patch(MODULE_PATH + '.Token.get_token_data', auto_spec=True, unsafe=True)
    def test_update_token_data_error(
        self,
        mock_get_token_data
    ):
        self.token.refresh_token = None
        self.token.created -= timedelta(121)
        with self.assertRaises(TokenExpiredError):
            self.token.update_token_data()


class TestCallbackRedirect(TestCase):
    def test_should_return_str(self):
        # given
        cb = CallbackRedirectFactory.build(session_key="abc", url="/green/alpha")
        # when/then
        self.assertEqual("abc: /green/alpha", str(cb))

    def test_should_return_repr(self):
        # given
        cb = CallbackRedirectFactory(session_key="abc", url="/green/alpha")
        # when/then
        expected = f"<CallbackRedirect(pk={cb.pk}): abc to /green/alpha>"
        self.assertEqual(expected, repr(cb))



"""unit tests for esi decorators"""

import logging
import os

from time import time
from unittest.mock import MagicMock, patch, Mock

from celery import Celery
from celery.contrib.testing.worker import start_worker
from celery.exceptions import Retry
from django.core.cache import cache
from django.contrib.auth.models import User
from django.contrib.auth.views import redirect_to_login
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test import TestCase, RequestFactory

from . import _generate_token, _store_as_Token
from ..rate_limiting import ESIRateLimitBucket, ESIRateLimits
from ..exceptions import ESIBucketLimitException, ESIErrorLimitException
from ..decorators import (
    _check_callback, esi_rate_limiter_bucketed, rate_limit_retry_task, tokens_required, token_required, single_use_token, wait_for_esi_errorlimit_reset
)
from ..models import Token, CallbackRedirect
SPEC_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "test_openapi.json"
)


logger = logging.getLogger(__name__)


class TestCheckCallback(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )

        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', 'xyz', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    def test_normal(self):
        logger.debug('start')
        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='xyz',
            token=self.token
        )
        response = _check_callback(request)
        self.assertEqual(response, self.token)

    def test_no_cb(self):
        logger.debug('start')
        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = _check_callback(request)
        self.assertIsNone(response)

    def test_no_token(self):
        logger.debug('start')
        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='xyz'
        )

        response = _check_callback(request)
        self.assertIsNone(response)

    def test_no_session(self):
        logger.debug('start')
        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)

        response = _check_callback(request)
        self.assertIsNone(response)


class TestTokensRequired(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )

        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    def test_token_already_exists(self):
        logger.debug('start')

        @tokens_required(scopes='abc')
        def my_view(request, tokens):
            return tokens

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response.first(),
            self.token
        )

    def test_token_and_cb_already_exists(self):
        logger.debug('start')

        @tokens_required(scopes='abc')
        def my_view(request, tokens):
            return tokens

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='xyz',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response.first(),
            self.token
        )

    @patch('esi.views.sso_redirect', autospec=True)
    def test_token_does_not_exist(self, mock_sso_redirect):
        logger.debug('start')

        @tokens_required(scopes='xyz')
        def my_view(request, tokens):
            return tokens

        mock_sso_redirect.return_value = 'sso_redirect_view_called'

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response,
            'sso_redirect_view_called'
        )

    def test_coming_back_from_sso(self):
        logger.debug('start')

        @tokens_required(scopes='abc', new=True)
        def my_view(request, tokens):
            return tokens

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='xyz',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response.first(),
            self.token
        )

    def test_user_not_authed(self):
        logger.debug('start')

        @tokens_required(scopes='abc')
        def my_view(request, tokens):
            return tokens

        mock_user = Mock(spec=User)
        mock_user.is_authenticated = False

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = mock_user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response.url,
            redirect_to_login(request.get_full_path()).url
        )


class TestTokenRequired(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )

        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    @patch('esi.views.select_token', autospec=True)
    @patch('esi.views.sso_redirect', autospec=True)
    def test_initial_call_with_matching_tokens(
        self,
        mock_sso_redirect,
        mock_select_token
    ):
        logger.debug('start')

        @token_required(scopes='abc')
        def my_view(request, tokens):
            return tokens

        mock_sso_redirect.return_value = 'sso_redirect_view_called'
        mock_select_token.return_value = 'select_token_view_called'

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response,
            'select_token_view_called'
        )

    @patch('esi.views.select_token', autospec=True)
    @patch('esi.views.sso_redirect', autospec=True)
    def test_initial_call_wo_matching_tokens(
        self,
        mock_sso_redirect,
        mock_select_token
    ):
        logger.debug('start')

        @token_required(scopes='xyz')
        def my_view(request, tokens):
            return tokens

        mock_sso_redirect.return_value = 'sso_redirect_view_called'
        mock_select_token.return_value = 'select_token_view_called'

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response,
            'sso_redirect_view_called'
        )

    def test_coming_back_from_sso_normal(
        self,
    ):
        logger.debug('start')

        @token_required(scopes='abc', new=True)
        def my_view(request, token):
            return token

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='qwe123',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response,
            self.token
        )

    @patch('esi.views.select_token', autospec=True)
    @patch('esi.views.sso_redirect', autospec=True)
    def test_user_wants_to_add_new_token(
        self,
        mock_sso_redirect,
        mock_select_token
    ):
        logger.debug('start')

        @token_required(scopes='abc')
        def my_view(request, token):
            return token

        mock_sso_redirect.return_value = 'sso_redirect_view_called'
        mock_select_token.return_value = 'select_token_view_called'

        request = self.factory.post(
            'https://www.example.com/my_view/',
            {
                '_add': False
            }
        )
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='qwe123',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response,
            self.token
        )

    @patch('esi.views.select_token', autospec=True)
    @patch('esi.views.sso_redirect', autospec=True)
    def test_user_chose_existing_token(
        self,
        mock_sso_redirect,
        mock_select_token
    ):
        logger.debug('start')

        @token_required(scopes='abc')
        def my_view(request, token):
            return token

        mock_sso_redirect.return_value = 'sso_redirect_view_called'
        mock_select_token.return_value = 'select_token_view_called'

        request = self.factory.post(
            'https://www.example.com/my_view/',
            {
                '_token': self.token.pk
            }
        )
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='qwe123',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response,
            self.token
        )

    @patch('esi.views.select_token', autospec=True)
    @patch('esi.views.sso_redirect', autospec=True)
    def test_user_chose_token_which_does_not_exist(
        self,
        mock_sso_redirect,
        mock_select_token
    ):
        logger.debug('start')

        @token_required(scopes='abc')
        def my_view(request, token):
            return token

        mock_sso_redirect.return_value = 'sso_redirect_view_called'
        mock_select_token.return_value = 'select_token_view_called'

        invalid_token_pk = max(int(x.pk) for x in Token.objects.all()) + 1
        request = self.factory.post(
            'https://www.example.com/my_view/',
            {
                '_token': invalid_token_pk
            }
        )
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response,
            'select_token_view_called'
        )


class TestSingleUseTokenRequired(TestCase):

    def setUp(self):
        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name="No Auth Character",
                scopes=['abc', '123']
            ),
            None
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    @patch('esi.views.sso_redirect', autospec=True)
    def test_initial_call_wo_matching_tokens(
        self,
        mock_sso_redirect
    ):
        logger.debug('start')

        @single_use_token(scopes='xyz')
        def my_view(request, tokens):
            return tokens

        mock_sso_redirect.return_value = 'sso_redirect_view_called'

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = None
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        response = my_view(request)
        self.assertEqual(
            response,
            'sso_redirect_view_called'
        )

    def test_coming_back_from_sso_normal(
        self,
    ):
        logger.debug('start')

        @single_use_token(scopes='abc', new=True)
        def my_view(request, token):
            return token

        request = self.factory.get('https://www.example.com/my_view/')
        request.user = None
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            url='https://www.example.com/redirect/',
            state='qwe123',
            token=self.token
        )

        response = my_view(request)
        self.assertEqual(
            response,
            self.token
        )


class TestESIRateLimitDecorator(TestCase):

    def setUp(self):
        self.bucket = ESIRateLimitBucket(
            "test-bucket",
            1,
            5
        )
        cache.clear()

    def test_raise(self):
        @esi_rate_limiter_bucketed(bucket=self.bucket)
        def my_func():
            return "Pass"

        _t = time()
        my_func()
        self.assertLess(time() - _t, 1)
        _t = time()
        with self.assertRaises(ESIBucketLimitException):
            my_func()

    def test_sleep(self):
        @esi_rate_limiter_bucketed(bucket=self.bucket, raise_on_limit=False)
        def my_func():
            return "Pass"

        _t = time()
        my_func()
        self.assertLess(time() - _t, 1)
        _t = time()
        my_func()
        duration = time() - _t
        self.assertGreater(duration, 5)


class TestESIErrorLimitDecorator(TestCase):

    def setUp(self):
        cache.clear()

    def test_sleep(self):
        @wait_for_esi_errorlimit_reset()
        def my_func():
            return "Pass"

        _t = time()
        my_func()
        self.assertLess(time() - _t, 1)
        cache.set("esi_error_limit_reset", 5, timeout=5)
        _t = time()
        my_func()
        duration = time() - _t
        self.assertGreater(duration, 5)

class TestESIErrorLimitDecorator(TestCase):
    def setUp(self):
        self.celery_app = Celery("mysite", broker_url="memory://localhost/")
        self.celery_app.conf.update(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
        self.celery_app.config_from_object("django.conf:settings")


    def test_no_limit_bucket(self):
        @self.celery_app.task(bind=True)
        @rate_limit_retry_task
        def _task(self):
            bucket = ESIRateLimitBucket(
                "test",
                1,
                5
            )
            raise ESIBucketLimitException(
                bucket=bucket,
                reset=2
            )

        with start_worker(
            self.celery_app,
            pool="solo",
            loglevel="info",
            perform_ping_check=False,
            shutdown_timeout=30
        ):
            with self.assertRaises(Retry):
                _task.delay()

    def test_error_global(self):
        @self.celery_app.task(bind=True)
        @rate_limit_retry_task
        def _task(self):
            raise ESIErrorLimitException(
                reset=2
            )

        with start_worker(
            self.celery_app,
            pool="solo",
            loglevel="info",
            perform_ping_check=False,
            shutdown_timeout=30
        ):
            with self.assertRaises(Retry):
                _task.delay()



import socket
from django.test import TestCase


def _dt_eveformat(dt: object) -> str:
    """converts a datetime to a string in eve format
    e.g. '2019-06-25T19:04:44'
    """
    from datetime import datetime

    dt2 = datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)
    return dt2.isoformat()


def _generate_token(
    character_id: int,
    character_name: str,
    access_token: str = 'access_token',
    refresh_token: str = 'refresh_token',
    character_owner_hash: str = 'character_owner_hash',
    scopes=None,  # list or string from ccp
    timestamp_dt: object = None,
    expires_in: int = 1200,
    sso_version: int = 2
) -> dict:
    import datetime as dt

    if timestamp_dt is None:
        timestamp_dt = dt.datetime.now(dt.timezone.utc)
    if scopes is None:
        scopes = [
            'esi-mail.read_mail.v1',
            'esi-wallet.read_character_wallet.v1',
            'esi-universe.read_structures.v1'
        ]
    token = {
        'access_token': access_token,
        'expires_in': expires_in,
        'refresh_token': refresh_token,
        'timestamp': int(timestamp_dt.timestamp()),
        'character_id': character_id,
        'name': character_name,
        'ExpiresOn': _dt_eveformat(timestamp_dt + dt.timedelta(seconds=expires_in)),
        'scp': scopes,
        'token_type': 'character',
        'owner': character_owner_hash,
        'IntellectualProperty': 'EVE',
        'sso_version': sso_version
    }
    return token


def _store_as_Token(token: dict, user: object) -> object:
    """Stores a generated token dict as Token object for given user

    returns Token object
    """
    from ..models import Scope, Token

    obj = Token.objects.create(
        access_token=token['access_token'],
        refresh_token=token['refresh_token'],
        user=user,
        character_id=token['character_id'],
        character_name=token['name'],
        token_type=token['token_type'],
        character_owner_hash=token['owner'],
        sso_version=token['sso_version']
    )

    if isinstance(token['scp'], str):
        token['scp'] = [token['scp']]

    for scope_name in token['scp']:
        scope, _ = Scope.objects.get_or_create(
            name=scope_name
        )
        obj.scopes.add(scope)

    return obj


class SocketAccessError(Exception):
    """Error raised when a test script accesses the network"""


class GuardedSocket(socket.socket):
    """A socket subclass that only allows loopback/localhost."""

    def _address_is_loopback(self, address):

        host = None
        if isinstance(address, tuple):
            host = address[0]
        else:
            host = address

        if isinstance(host, bytes):
            host = host.decode()

        # quick allow by obvious names
        if host in ("localhost", "127.0.0.1", "::1"):
            return True
        else:
            return False

    def connect(self, address):
        if not self._address_is_loopback(address):
            raise SocketAccessError(f"Attempt to connect to non-localhost address: {address!r}")
        return super().connect(address)


class NoSocketsTestCase(TestCase):
    """Variation of Django's TestCase class that prevents any network use.

    Example:

        .. code-block:: python

            class TestMyStuff(NoSocketsTestCase):
                def test_should_do_what_i_need(self):
                    ...

    """
    @classmethod
    def setUpClass(cls):
        cls._socket_original = socket.socket
        socket.socket = GuardedSocket
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls._socket_original
        super().tearDownClass()



"""unit tests for esi"""

from django.test import TestCase
from ..templatetags.scope_tags import scope_friendly_name


class TestScope(TestCase):
    """tests for template tag"""

    def test_friendly_name_fail(self):
        x = scope_friendly_name('dummy_scope')
        self.assertEqual(
            "dummy_scope",
            x
        )

    def test_friendly_name_pass(self):
        x = scope_friendly_name('test.dummy_scope.test')
        self.assertEqual(
            "dummy scope",
            x
        )



# flake8: noqa
"""script for testing client with live requests to ESI

Run this script directly. Make sure to also set the environment variable
DJANGO_PROJECT_PATH and DJANGO_SETTINGS_MODULE to match your setup:

You can see the result in your main log file of your Django installation.

Example:
export DJANGO_PROJECT_PATH="/home/erik997/dev/python/django-dev/mysite"
export DJANGO_SETTINGS_MODULE="mysite.settings"

"""

# start django project
import os
import sys

if not 'DJANGO_PROJECT_PATH' in os.environ:
    print('DJANGO_PROJECT_PATH is not set')
    exit(1)

if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print('DJANGO_SETTINGS_MODULE is not set')
    exit(1)

sys.path.insert(0, os.environ['DJANGO_PROJECT_PATH'])
import django
django.setup()

# normal imports
import concurrent.futures
import logging

from django.core.cache import cache
from esi.clients import EsiClientProvider

cache.clear()
esi = EsiClientProvider()

logger = logging.getLogger('__name__')
logger.level = logging.DEBUG


def main():
    # esi.client.Universe.get_universe_types().results()
    o = esi.client.Universe.get_universe_types_type_id(type_id=34562)
    o.results()

if __name__ == '__main__':
    print('Script started...')
    main()
    print('DONE')



"""Factory for generating Eve Onlie conformant JWTs for testing."""

import datetime as dt

from jose import jwt
from pytz import utc

_RSA_PRIVATE_KEY = """
-----BEGIN PRIVATE KEY-----
MIIEvwIBADANBgkqhkiG9w0BAQEFAASCBKkwggSlAgEAAoIBAQC7VJTUt9Us8cKj
MzEfYyjiWA4R4/M2bS1GB4t7NXp98C3SC6dVMvDuictGeurT8jNbvJZHtCSuYEvu
NMoSfm76oqFvAp8Gy0iz5sxjZmSnXyCdPEovGhLa0VzMaQ8s+CLOyS56YyCFGeJZ
qgtzJ6GR3eqoYSW9b9UMvkBpZODSctWSNGj3P7jRFDO5VoTwCQAWbFnOjDfH5Ulg
p2PKSQnSJP3AJLQNFNe7br1XbrhV//eO+t51mIpGSDCUv3E0DDFcWDTH9cXDTTlR
ZVEiR2BwpZOOkE/Z0/BVnhZYL71oZV34bKfWjQIt6V/isSMahdsAASACp4ZTGtwi
VuNd9tybAgMBAAECggEBAKTmjaS6tkK8BlPXClTQ2vpz/N6uxDeS35mXpqasqskV
laAidgg/sWqpjXDbXr93otIMLlWsM+X0CqMDgSXKejLS2jx4GDjI1ZTXg++0AMJ8
sJ74pWzVDOfmCEQ/7wXs3+cbnXhKriO8Z036q92Qc1+N87SI38nkGa0ABH9CN83H
mQqt4fB7UdHzuIRe/me2PGhIq5ZBzj6h3BpoPGzEP+x3l9YmK8t/1cN0pqI+dQwY
dgfGjackLu/2qH80MCF7IyQaseZUOJyKrCLtSD/Iixv/hzDEUPfOCjFDgTpzf3cw
ta8+oE4wHCo1iI1/4TlPkwmXx4qSXtmw4aQPz7IDQvECgYEA8KNThCO2gsC2I9PQ
DM/8Cw0O983WCDY+oi+7JPiNAJwv5DYBqEZB1QYdj06YD16XlC/HAZMsMku1na2T
N0driwenQQWzoev3g2S7gRDoS/FCJSI3jJ+kjgtaA7Qmzlgk1TxODN+G1H91HW7t
0l7VnL27IWyYo2qRRK3jzxqUiPUCgYEAx0oQs2reBQGMVZnApD1jeq7n4MvNLcPv
t8b/eU9iUv6Y4Mj0Suo/AU8lYZXm8ubbqAlwz2VSVunD2tOplHyMUrtCtObAfVDU
AhCndKaA9gApgfb3xw1IKbuQ1u4IF1FJl3VtumfQn//LiH1B3rXhcdyo3/vIttEk
48RakUKClU8CgYEAzV7W3COOlDDcQd935DdtKBFRAPRPAlspQUnzMi5eSHMD/ISL
DY5IiQHbIH83D4bvXq0X7qQoSBSNP7Dvv3HYuqMhf0DaegrlBuJllFVVq9qPVRnK
xt1Il2HgxOBvbhOT+9in1BzA+YJ99UzC85O0Qz06A+CmtHEy4aZ2kj5hHjECgYEA
mNS4+A8Fkss8Js1RieK2LniBxMgmYml3pfVLKGnzmng7H2+cwPLhPIzIuwytXywh
2bzbsYEfYx3EoEVgMEpPhoarQnYPukrJO4gwE2o5Te6T5mJSZGlQJQj9q4ZB2Dfz
et6INsK0oG8XVGXSpQvQh3RUYekCZQkBBFcpqWpbIEsCgYAnM3DQf3FJoSnXaMhr
VBIovic5l0xFkEHskAjFTevO86Fsz1C2aSeRKSqGFoOQ0tmJzBEs1R6KqnHInicD
TQrKhArgLXX4v3CddjfTRJkFWDbE/CkvKZNOrcf1nhaGCPspRJj2KUkj1Fhl9Cnc
dn/RsYEONbwQSjIfMPkvxF+8HQ==
-----END PRIVATE KEY-----
"""

_RSA_PUBLIC_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAu1SU1LfVLPHCozMxH2Mo
4lgOEePzNm0tRgeLezV6ffAt0gunVTLw7onLRnrq0/IzW7yWR7QkrmBL7jTKEn5u
+qKhbwKfBstIs+bMY2Zkp18gnTxKLxoS2tFczGkPLPgizskuemMghRniWaoLcyeh
kd3qqGElvW/VDL5AaWTg0nLVkjRo9z+40RQzuVaE8AkAFmxZzow3x+VJYKdjykkJ
0iT9wCS0DRTXu269V264Vf/3jvredZiKRkgwlL9xNAwxXFg0x/XFw005UWVRIkdg
cKWTjpBP2dPwVZ4WWC+9aGVd+Gyn1o0CLelf4rEjGoXbAAEgAqeGUxrcIlbjXfbc
mwIDAQAB
-----END PUBLIC KEY-----
"""
_JWK_PUBLIC_KEY = """u1SU1LfVLPHCozMxH2Mo4lgOEePzNm0tRgeLezV6ffAt0gunVTLw7onLRnrq0_IzW7yWR7QkrmBL7jTKEn5u-qKhbwKfBstIs-bMY2Zkp18gnTxKLxoS2tFczGkPLPgizskuemMghRniWaoLcyehkd3qqGElvW_VDL5AaWTg0nLVkjRo9z-40RQzuVaE8AkAFmxZzow3x-VJYKdjykkJ0iT9wCS0DRTXu269V264Vf_3jvredZiKRkgwlL9xNAwxXFg0x_XFw005UWVRIkdgcKWTjpBP2dPwVZ4WWC-9aGVd-Gyn1o0CLelf4rEjGoXbAAEgAqeGUxrcIlbjXfbcmw"""  # noqa: E501
_ALGORITHM = jwt.ALGORITHMS.RS256
_AUDIENCE = "EVE Online"
_HEADERS = {"alg": "RS256", "kid": "JWT-Signature-Key", "typ": "JWT"}
_ISSUER = "login.eveonline.com"


def generate_jwk():
    return {
        "alg": _ALGORITHM,
        "e": "AQAB",
        "kid": "JWT-Signature-Key",
        "kty": "RSA",
        "n": _JWK_PUBLIC_KEY,
        "use": "sig",
    }


def _generate_claims(
    issued_at: dt.datetime,
    character_id: int,
    character_name: str,
    issuer=None,
    audience=None,
) -> dict:
    if not issuer:
        issuer = _ISSUER
    expires_at = issued_at + dt.timedelta(minutes=20)
    claims = {
        "scp": "esi-characters.read_medals.v1",
        "jti": "xxx",
        "kid": "JWT-Signature-Key",
        "sub": f"CHARACTER:EVE:{character_id}",
        "azp": "yyy",
        "tenant": "tranquility",
        "tier": "live",
        "region": "world",
        "name": str(character_name),
        "owner": "OWNER-HASH",
        "exp": int(expires_at.timestamp()),
        "iat": int(issued_at.timestamp()),
        "iss": _ISSUER,
    }
    if audience is None:
        audience = _AUDIENCE
    if audience:
        claims["aud"] = audience
    return claims


def generate_token(
    character_id: int,
    character_name: str,
    issued_at: dt.datetime = None,
    issuer=None,
    audience=None,
) -> tuple[str, dict]:
    """Generate a JWT for Eve Online."""
    if not issued_at:
        issued_at = dt.datetime.now(tz=utc)
    claims = _generate_claims(
        issued_at=issued_at,
        character_id=character_id,
        character_name=character_name,
        issuer=issuer,
        audience=audience,
    )
    token_string = jwt.encode(
        claims=claims, key=_RSA_PRIVATE_KEY, algorithm=_ALGORITHM, headers=_HEADERS
    )
    return token_string, claims


def validate():
    issued_at = dt.datetime.now(tz=utc)
    token, claims = generate_token(
        issued_at=issued_at, character_id=1001, character_name="Bruce Wayne"
    )
    new_claims = jwt.decode(
        token=token, key=generate_jwk(), audience=_AUDIENCE, issuer=_ISSUER
    )
    print("Validate successful: ", claims == new_claims)


if __name__ == "__main__":
    validate()



from typing import Any
from django.db.models import QuerySet, Manager
from .models import Token


class TokenQueryset(QuerySet[Token]):
    def get_expired(self) -> "TokenQueryset":
        ...

    def bulk_refresh(self) -> "TokenQueryset":
        ...

    def require_scopes(self, scope_string: str | list) -> "TokenQueryset":
        ...

    def require_scopes_exact(self, scope_string: str | list) -> "TokenQueryset":
        ...

    def require_valid(self) -> "TokenQueryset":
        ...

    def equivalent_to(self, token: Token) -> "TokenQueryset":
        ...

    def filter(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def exclude(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def order_by(self, *field_names: str) -> "TokenQueryset":
        ...

    def annotate(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def select_related(self, *fields: str) -> "TokenQueryset":
        ...

    def prefetch_related(self, *lookups: str) -> "TokenQueryset":
        ...

    def distinct(self, *field_names: str) -> "TokenQueryset":
        ...

    def none(self) -> "TokenQueryset":
        ...

    def all(self) -> "TokenQueryset":
        ...


class TokenManager(Manager[Token]):
    def get_queryset(self) -> "TokenQueryset":
        ...

    def require_scopes(self, scope_string: str | list) -> "TokenQueryset":
        ...

    def filter(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def exclude(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def order_by(self, *field_names: str) -> "TokenQueryset":
        ...

    def annotate(self, *args: Any, **kwargs: Any) -> "TokenQueryset":
        ...

    def select_related(self, *fields: str) -> "TokenQueryset":
        ...

    def prefetch_related(self, *lookups: str) -> "TokenQueryset":
        ...

    def distinct(self, *field_names: str) -> "TokenQueryset":
        ...

    def none(self) -> "TokenQueryset":
        ...

    def all(self) -> "TokenQueryset":
        ...



from esi.models import Token
from string import capwords


def get_token(character_id: int, scopes: list) -> Token:
    """Helper method to get a valid token for a specific character with specific scopes.

    Args:
        character_id: Character to filter on.
        scopes: array of ESI scope strings to search for.

    Returns:
        Matching Token
    """
    qs = (
        Token.objects
        .filter(character_id=character_id)
        .require_scopes(scopes)
        .require_valid()
    )
    token = qs.first()
    if token is None:
        raise Token.DoesNotExist(
            f"No valid token found for character_id={character_id} with required scopes."
        )
    return token


def pascal_case_string(string: str) -> str:
    """
    Convert a string to PascalCase by capitalizing the first letter of each word and removing spaces,
    but only if the string contains spaces or hyphens.

    This function checks if the input string contains spaces or hyphens. If so, it replaces hyphens with spaces,
    capitalizes the first letter of each word, removes the spaces, and returns the resulting PascalCase string.
    If the input string does not contain spaces or hyphens, it is returned unchanged.

    Behaviour:
    Any string containing spaces or hyphens will be converted to PascalCase.
    Strings without spaces or hyphens will be returned unchanged.
    This gives you the opportunity to use already formatted strings as needed.

    Examples:
    - "app name" -> "AppName"
    - "app-name" -> "AppName"
    - "appname" -> "appname"
    - "AppName" -> "AppName"
    - "appName" -> "appName"
    - "app_name" -> "app_name"

    :param string: The input string to be converted to PascalCase.
    :type string: str
    :return: The PascalCase formatted string, or the original string if no spaces or hyphens are present.
    :rtype: str
    """

    # Check if the string contains spaces or hyphens
    if any(c in string for c in (" ", "-")):
        # Replace hyphens with spaces, capitalize each word, and remove spaces
        return capwords(string.replace("-", " ")).replace(" ", "")

    # Return the original string if no spaces or hyphens are present
    return string



from datetime import timedelta
from unittest.mock import Mock, patch

import requests_mock
from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test import RequestFactory, TestCase
from django.utils.timezone import now

from ..errors import IncompleteResponseError, TokenError
from ..managers import _process_scopes
from ..models import Token
from . import _generate_token, _store_as_Token
from .jwt_factory import generate_jwk, generate_token


class TestProcessScopes(TestCase):
    def test_none(self):
        self.assertSetEqual(_process_scopes(None), set())

    def test_empty_list(self):
        self.assertSetEqual(_process_scopes([]), set())

    def test_single_string_1(self):
        self.assertSetEqual(_process_scopes(["one"]), {"one"})

    def test_single_string_2(self):
        self.assertSetEqual(_process_scopes(["one two three"]), {"one", "two", "three"})

    def test_list(self):
        self.assertSetEqual(
            _process_scopes(["one", "two", "three"]), {"one", "two", "three"}
        )

    def test_tuple(self):
        self.assertSetEqual(
            _process_scopes(("one", "two", "three")), {"one", "two", "three"}
        )


class TestTokenGetExpired(TestCase):
    @patch("esi.models.app_settings.ESI_TOKEN_VALID_DURATION", 120)
    def test_get_expired(self):
        self.user1 = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )
        self.user2 = User.objects.create_user(
            "Peter Parker", "abc@example.com", "password"
        )
        _store_as_Token(
            _generate_token(
                character_id=101, character_name=self.user1.username, scopes=["abc"]
            ),
            self.user1,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=102, character_name=self.user2.username, scopes=["xyz"]
            ),
            self.user2,
        )
        self.assertEqual(list(Token.objects.get_queryset().get_expired()), [])

        t2.created -= timedelta(121)
        t2.save()
        self.assertEqual(list(Token.objects.get_queryset().get_expired()), [t2])


@patch("esi.managers.app_settings.ESI_SSO_CLIENT_ID", "abc")
@patch("esi.managers.app_settings.ESI_SSO_CLIENT_SECRET", "xyz")
@patch("esi.models.Token.delete", autospec=True)
@patch("esi.models.Token.refresh", autospec=True)
@patch("esi.managers.requests.auth.HTTPBasicAuth", autospec=True)
@patch("esi.managers.OAuth2Session", autospec=True)
class TestTokenBulkRefresh(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )

    def test_bulk_refresh_normal(
        self,
        mock_OAuth2Session,
        mock_HTTPBasicAuth,
        mock_Token_refresh,
        mock_Token_delete,
    ):
        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        incomplete_qs = Token.objects.get_queryset().bulk_refresh()
        self.assertEqual(mock_Token_refresh.call_count, 2)
        self.assertEqual(mock_Token_delete.call_count, 0)
        self.assertSetEqual(set(incomplete_qs), {t1, t2})

        # Note
        # looks like a bug in bulk_refresh():
        # this filter can never find anything, because refresh_token
        # can not be null:
        #   self.filter(refresh_token__isnull=True).get_expired().delete()

    def test_bulk_refresh_token_error(
        self,
        mock_OAuth2Session,
        mock_HTTPBasicAuth,
        mock_Token_refresh,
        mock_Token_delete,
    ):
        mock_Token_refresh.side_effect = TokenError

        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        incomplete_qs = Token.objects.get_queryset().bulk_refresh()
        self.assertEqual(mock_Token_refresh.call_count, 2)
        self.assertEqual(mock_Token_delete.call_count, 2)
        self.assertSetEqual(set(incomplete_qs), {t1, t2})

    def test_bulk_refresh_incomplete_response_error(
        self,
        mock_OAuth2Session,
        mock_HTTPBasicAuth,
        mock_Token_refresh,
        mock_Token_delete,
    ):
        mock_Token_refresh.side_effect = IncompleteResponseError

        character_id = 99
        character_name = "Bruce Wayne"
        _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        incomplete_qs = Token.objects.get_queryset().bulk_refresh()
        self.assertEqual(mock_Token_refresh.call_count, 2)
        self.assertEqual(mock_Token_delete.call_count, 0)
        self.assertSetEqual(set(incomplete_qs), set())


@patch("esi.models.app_settings.ESI_TOKEN_VALID_DURATION", 120)
@patch("esi.managers.TokenQueryset.bulk_refresh", autospec=True)
class TestTokenRequireValid(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )

    def test_require_valid_none_expired(self, mock_bulk_refresh):
        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        mock_bulk_refresh.return_value = Token.objects.none()

        self.assertSetEqual(set(Token.objects.get_queryset().require_valid()), {t1, t2})

    def test_require_valid_some_expired(self, mock_bulk_refresh):
        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        t2.created -= timedelta(121)
        t2.save()
        mock_bulk_refresh.return_value = Token.objects.filter(pk__in=[t2.pk])

        self.assertSetEqual(set(Token.objects.get_queryset().require_valid()), {t1, t2})

    def test_require_valid_one_refresh_error(self, mock_bulk_refresh):
        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["xyz"]
            ),
            self.user,
        )
        t2.created -= timedelta(121)
        t2.save()
        mock_bulk_refresh.return_value = Token.objects.none()

        self.assertSetEqual(set(Token.objects.get_queryset().require_valid()), {t1})


@patch("esi.models.app_settings.ESI_TOKEN_VALID_DURATION", 120)
@patch("esi.models.Token.refresh", spec=True)
class TestTokenBulkRefreshXRequireValid(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )

    def test_require_valid_all_expired(self, mock_token_refresh):
        """Test demonstrates the bug introduced with 5.3.0b1.

        bulk_refresh() does not return the token after it has been refreshed.
        Then also require_valid() does not return the refreshed token.
        Apps calling require_valid() will then assume there is no valid token.
        The previous tests did not catch this bug, because it they were testing
        the methods one by one and mocking out responses from other methods.
        """
        def my_refresh(*args, **kwargs):
            t.created = now()
            t.save()

        # given
        mock_token_refresh.side_effect = my_refresh
        character_id = 99
        character_name = "Bruce Wayne"
        t: Token = _store_as_Token(
            _generate_token(
                character_id=character_id, character_name=character_name, scopes=["abc"]
            ),
            self.user,
        )
        t.created -= timedelta(121)
        t.save()

        # when
        qs = Token.objects.get_queryset().require_valid()

        # then
        expected = set(qs.values_list("pk", flat=True))
        self.assertSetEqual(expected, {t.pk})


class TestTokenRequireScopes(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )

    def test_require_scopes_normal(self):
        character_id = 99
        character_name = "Bruce Wayne"
        t1 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz", "123"],
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz"],
            ),
            self.user,
        )
        t3 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "123"],
            ),
            self.user,
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes("abc")), {t1, t2, t3}
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes("xyz")), {t1, t2}
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes("123")), {t1, t3}
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes("555")), set()
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes("abc xyz 123")), {t1}
        )

    def test_require_scopes_empty(self):
        character_id = 99
        character_name = "Bruce Wayne"
        _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz", "123"],
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(character_id=character_id, character_name=character_name),
            self.user,
        )
        t2.scopes.all().delete()
        _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "123"],
            ),
            self.user,
        )
        self.assertSetEqual(set(Token.objects.get_queryset().require_scopes("")), {t2})

    def test_require_scopes_exact(self):
        character_id = 99
        character_name = "Bruce Wayne"
        _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz", "123"],
            ),
            self.user,
        )
        t2 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz"],
            ),
            self.user,
        )
        t3 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "123"],
            ),
            self.user,
        )
        t4 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "123"],
            ),
            self.user,
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes_exact("abc")), set()
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes_exact("abc xyz")), {t2}
        )
        self.assertSetEqual(
            set(Token.objects.get_queryset().require_scopes_exact("abc 123")), {t3, t4}
        )

    def test_require_scopes_exact_2(self):
        character_id = 99
        character_name = "Bruce Wayne"
        _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz", "123"],
            ),
            self.user,
        )
        _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["abc", "xyz"],
            ),
            self.user,
        )
        t3 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["xyz", "123"],
            ),
            self.user,
        )
        t4 = _store_as_Token(
            _generate_token(
                character_id=character_id,
                character_name=character_name,
                scopes=["xyz", "123"],
            ),
            self.user,
        )
        self.assertSetEqual(set(Token.objects.get_queryset().equivalent_to(t3)), {t4})


class TestTokenManager(TestCase):

    def setUp(self):
        self.factory = RequestFactory()

        self.user1 = User.objects.create_user(
            "Bruce Wayne", "abc@example.com", "password"
        )

    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_ID", "abc")
    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_SECRET", "xyz")
    @patch("esi.managers.app_settings.ESI_SSO_CALLBACK_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_VERIFY_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_ALWAYS_CREATE_TOKEN", False)
    @patch("esi.managers.TokenManager._decode_jwt")
    @patch("esi.managers.OAuth2Session", autospec=True)
    def test_create_from_code_single_scope(self, mock_OAuth2Session, mock_decode_jwt):
        """Normal case with refresh token"""
        mock_oauth = Mock()
        mock_oauth.request.return_value.json.return_value = _generate_token(
            99, "Bruce Wayne", scopes="publicData"
        )
        mock_oauth.fetch_token.return_value = {
            "access_token": "access_token",
            "refresh_token": "refresh_token",
            "token_type": "Bearer",
            "expires_in": 1200,
        }
        mock_OAuth2Session.return_value = mock_oauth
        mock_decode_jwt.return_value = _generate_token(
            99, "Bruce Wayne", scopes="publicData"
        )
        # create new token from code
        token1 = Token.objects.create_from_code("abc123xyz")
        self.assertEqual(token1.character_id, 99)
        self.assertEqual(token1.character_name, "Bruce Wayne")
        self.assertEqual(token1.scopes.all().count(), 1)
        # should return existing token instead of creating a new one
        # since ESI_ALWAYS_CREATE_TOKEN is False
        token2 = Token.objects.create_from_code("11abc123xyz")
        self.assertEqual(token1, token2)

    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_ID", "abc")
    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_SECRET", "xyz")
    @patch("esi.managers.app_settings.ESI_SSO_CALLBACK_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_VERIFY_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_ALWAYS_CREATE_TOKEN", False)
    @patch("esi.managers.TokenManager._decode_jwt")
    @patch("esi.managers.OAuth2Session", autospec=True)
    def test_create_from_code_1(self, mock_OAuth2Session, mock_decode_jwt):
        """Normal case with refresh token"""
        mock_oauth = Mock()
        mock_oauth.request.return_value.json.return_value = _generate_token(
            99,
            "Bruce Wayne",
            scopes=[
                "esi-calendar.read_calendar_events.v1",
                "esi-location.read_location.v1",
                "esi-location.read_ship_type.v1",
                "esi-unknown-scope",
            ],
        )
        mock_oauth.fetch_token.return_value = {
            "access_token": "access_token",
            "refresh_token": "refresh_token",
            "token_type": "Bearer",
            "expires_in": 1200,
        }
        mock_OAuth2Session.return_value = mock_oauth
        mock_decode_jwt.return_value = _generate_token(
            99,
            "Bruce Wayne",
            scopes=[
                "esi-calendar.read_calendar_events.v1",
                "esi-location.read_location.v1",
                "esi-location.read_ship_type.v1",
                "esi-unknown-scope",
            ],
        )
        # create new token from code
        token1 = Token.objects.create_from_code("abc123xyz")
        self.assertEqual(token1.character_id, 99)
        self.assertEqual(token1.character_name, "Bruce Wayne")
        self.assertEqual(token1.scopes.all().count(), 4)

        # should return existing token instead of creating a new one
        # since ESI_ALWAYS_CREATE_TOKEN is False
        token2 = Token.objects.create_from_code("11abc123xyz")
        self.assertEqual(token1, token2)

    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_ID", "abc")
    @patch("esi.managers.app_settings.ESI_SSO_CLIENT_SECRET", "xyz")
    @patch("esi.managers.app_settings.ESI_SSO_CALLBACK_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_TOKEN_VERIFY_URL", "localhost")
    @patch("esi.managers.app_settings.ESI_ALWAYS_CREATE_TOKEN", False)
    @patch("esi.managers.TokenManager._decode_jwt")
    @patch("esi.managers.OAuth2Session", autospec=True)
    def test_create_from_code_2(self, mock_OAuth2Session, mock_decode_jwt):
        """Special case w/o refresh token"""
        mock_oauth = Mock()
        mock_oauth.request.return_value.json.return_value = _generate_token(
            99,
            "Bruce Wayne",
            scopes=[
                "esi-calendar.read_calendar_events.v1",
                "esi-location.read_location.v1",
                "esi-location.read_ship_type.v1",
                "esi-unknown-scope",
            ],
        )
        mock_oauth.fetch_token.return_value = {
            "access_token": "access_token",
            "refresh_token": None,
            "token_type": "Bearer",
            "expires_in": 1200,
        }
        mock_OAuth2Session.return_value = mock_oauth
        mock_decode_jwt.return_value = _generate_token(
            99,
            "Bruce Wayne",
            scopes=[
                "esi-calendar.read_calendar_events.v1",
                "esi-location.read_location.v1",
                "esi-location.read_ship_type.v1",
                "esi-unknown-scope",
            ],
        )
        # create new token from code
        token1 = Token.objects.create_from_code("abc123xyz")
        self.assertEqual(token1.character_id, 99)
        self.assertEqual(token1.character_name, "Bruce Wayne")
        self.assertEqual(token1.scopes.all().count(), 4)

        # should return existing token instead of creating a new one
        # since ESI_ALWAYS_CREATE_TOKEN is False
        token2 = Token.objects.create_from_code("11abc123xyz")
        self.assertEqual(token1, token2)

    @patch("esi.managers.TokenManager.create_from_code", autospec=True)
    def test_create_from_request(self, mock_create_from_code):
        mock_create_from_code.return_value = "we got you"

        request = self.factory.get("https://www.example.com?code=abc123")
        request.user = self.user1

        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        x = Token.objects.create_from_request(request)
        self.assertEqual(x, "we got you")
        self.assertEqual(mock_create_from_code.call_args[0][1], "abc123")


@requests_mock.Mocker()
class TestTokenManagerValidateAccessToken(TestCase):
    def test_should_return_token_1(self, requests_mocker):
        # given
        jwks = {"keys": [generate_jwk()]}
        requests_mocker.register_uri(
            "GET", url="https://login.eveonline.com/oauth/jwks", json=jwks
        )
        access_token, _ = generate_token(1001, "Bruce Wayne")
        # when
        token = Token.objects.validate_access_token(access_token)
        # then
        self.assertEqual(token["character_id"], 1001)
        self.assertEqual(token["name"], "Bruce Wayne")
        self.assertEqual(token["token_type"], "character")

    def test_should_return_token_2(self, requests_mocker):
        # given
        jwks = {"keys": [generate_jwk()]}
        requests_mocker.register_uri(
            "GET", url="https://login.eveonline.com/oauth/jwks", json=jwks
        )
        access_token, _ = generate_token(
            1001, "Bruce Wayne", issuer="https://login.eveonline.com"
        )
        # when
        token = Token.objects.validate_access_token(access_token)
        # then
        self.assertEqual(token["character_id"], 1001)
        self.assertEqual(token["name"], "Bruce Wayne")
        self.assertEqual(token["token_type"], "character")

    def test_should_return_none_when_no_jwk(self, requests_mocker):
        # given
        jwks = dict()
        requests_mocker.register_uri(
            "GET", url="https://login.eveonline.com/oauth/jwks", json=jwks
        )
        access_token, _ = generate_token(1001, "Bruce Wayne", audience=False)
        # when
        result = Token.objects.validate_access_token(access_token)
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_expired(self, requests_mocker):
        # given
        jwks = {"keys": [generate_jwk()]}
        requests_mocker.register_uri(
            "GET", url="https://login.eveonline.com/oauth/jwks", json=jwks
        )
        issued_at = now() - timedelta(hours=3)
        access_token, _ = generate_token(1001, "Bruce Wayne", issued_at=issued_at)
        # when
        result = Token.objects.validate_access_token(access_token)
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_invalid(self, requests_mocker):
        # given
        jwks = {"keys": [generate_jwk()]}
        requests_mocker.register_uri(
            "GET", url="https://login.eveonline.com/oauth/jwks", json=jwks
        )
        access_token = "invalid"
        # when
        result = Token.objects.validate_access_token(access_token)
        # then
        self.assertIsNone(result)



# flake8: noqa
"""script for testing client with live requests to ESI

This script is doing an authed request and required at least one valid token to be stored in the database. To create a valid token you can simple run through
the esi test app once.

Run this script directly. Make sure to also set the environment variable
DJANGO_PROJECT_PATH and DJANGO_SETTINGS_MODULE to match your setup:

You can see the result in your main log file of your Django installation.

Example:
export DJANGO_PROJECT_PATH="/home/erik997/dev/python/django-dev/mysite"
export DJANGO_SETTINGS_MODULE="mysite.settings"

"""

# start django project
import os
import sys

if not 'DJANGO_PROJECT_PATH' in os.environ:
    print('DJANGO_PROJECT_PATH is not set')
    exit(1)

if not 'DJANGO_SETTINGS_MODULE' in os.environ:
    print('DJANGO_SETTINGS_MODULE is not set')
    exit(1)

sys.path.insert(0, os.environ['DJANGO_PROJECT_PATH'])
import django
django.setup()

# normal imports
import concurrent.futures
import logging

from django.core.cache import cache
from esi.models import Token
from esi.clients import EsiClientProvider

cache.clear()
esi = EsiClientProvider()

logger = logging.getLogger('__name__')
logger.level = logging.DEBUG


def main():
    token = Token.objects.filter(scopes__name='esi-characters.read_medals.v1').first()
    response = (
        esi.client.Character
        .get_characters_character_id_medals(
            character_id=token.character_id,
            token=token.valid_access_token()
        )
        .result()
    )

if __name__ == '__main__':
    print('Script started...')
    main()
    print('DONE')



"""Factories for test objects using factory boy."""

import factory
import factory.fuzzy

from esi.models import Token, Scope, CallbackRedirect


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "auth.User"
        django_get_or_create = ("username",)
        exclude = ("_generated_name",)

    _generated_name = factory.Faker("name")
    username = factory.LazyAttribute(lambda obj: obj._generated_name.replace(" ", "_"))
    first_name = factory.LazyAttribute(lambda obj: obj._generated_name.split(" ")[0])
    last_name = factory.LazyAttribute(lambda obj: obj._generated_name.split(" ")[1])
    email = factory.LazyAttribute(
        lambda obj: f"{obj.first_name.lower()}.{obj.last_name.lower()}@example.com"
    )


class TokenFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Token

    access_token = factory.faker.Faker("pystr", min_chars=2_507, max_chars=2_507)
    refresh_token = factory.faker.Faker("pystr", min_chars=24, max_chars=24)
    user = factory.SubFactory(UserFactory)
    character_id = factory.fuzzy.FuzzyInteger(low=90_000_000, high=98_000_000)
    character_name = factory.faker.Faker("name")
    token_type = "Character"
    character_owner_hash = factory.faker.Faker("pystr", min_chars=28, max_chars=28)
    sso_version = 2

    @factory.post_generation
    def scopes(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for scope in extracted:
                self.scopes.add(scope)


class ScopeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Scope

    name = factory.faker.Faker("word")
    help_text = factory.faker.Faker("sentence")


class CallbackRedirectFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = CallbackRedirect

    session_key = factory.faker.Faker("md5")
    url = factory.faker.Faker("uri_path", deep=3)
    state = factory.faker.Faker("sha1")



from django.conf import settings
from django.core.checks import Error, Tags, Warning, register


@register(Tags.security)
def check_sso_application_settings(*args, **kwargs):

    errors = []

    if (
        not hasattr(settings, "ESI_SSO_CLIENT_ID")
        or not hasattr(settings, "ESI_SSO_CLIENT_SECRET")
        or not hasattr(settings, "ESI_SSO_CALLBACK_URL")
    ):
        if settings.DEBUG:
            errors.append(
                Warning(
                    'ESI SSO application settings are not configured.',
                    hint='SSO features will not work.',
                    id='esi.W001'
                )
            )
        else:
            errors.append(
                Error(
                    'ESI SSO application settings cannot be blank.',
                    hint=(
                        'Register an application at https://developers.eveonline.com '
                        'and add ESI_SSO_CLIENT_ID, ESI_SSO_CLIENT_SECRET, and '
                        'ESI_SSO_CALLBACK_URL to your project settings.'
                    ),
                    id='esi.E001'
                )
            )

    # Check for ESI_USER_CONTACT_EMAIL
    if hasattr(settings, "ESI_USER_CONTACT_EMAIL"):
        # Check if ESI_USER_CONTACT_EMAIL is empty
        if settings.ESI_USER_CONTACT_EMAIL == "":
            errors.append(
                Error(
                    msg="'ESI_USER_CONTACT_EMAIL' is empty. A valid email is required as maintainer contact for CCP.",
                    hint="",
                    id="esi.E002",
                )
            )
    # ESI_USER_CONTACT_EMAIL not found
    else:
        errors.append(
            Error(
                msg="No 'ESI_USER_CONTACT_EMAIL' found is settings. A valid email is required as maintainer contact for CCP.",
                hint="",
                id="esi.E003",
            )
        )
    return errors



from unittest.mock import patch

from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
    Http404,
    HttpResponseBadRequest
)
from django.test import TestCase, RequestFactory

from . import _generate_token, _store_as_Token
from ..models import CallbackRedirect
from ..views import sso_redirect, receive_callback, select_token


ESI_SSO_CLIENT_ID = 'abc'
ESI_SSO_CALLBACK_URL = 'https://www.example.com/callback/'
ESI_OAUTH_LOGIN_URL = 'https://www.example.com/oauth/'

redirect_sub_url = '/%s/' % ('x' * (2048 - 25))
redirect_url = 'https://www.example.com' + redirect_sub_url


class TestSsoCallbackView(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )
        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', 'xyz', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    @patch('esi.views.app_settings.ESI_OAUTH_LOGIN_URL', ESI_OAUTH_LOGIN_URL)
    @patch('esi.views.OAuth2Session', autospec=True)
    def test_redirect_to_url_no_scopes(self, mock_OAuth2Session):
        state = 'my_awesome_state'
        mock_OAuth2Session.return_value.authorization_url.return_value = \
            (redirect_url, state)

        request = self.factory.get(redirect_url)
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = sso_redirect(request)

        self.assertTrue(mock_OAuth2Session.called)
        args, kwargs = mock_OAuth2Session.call_args
        self.assertEqual(kwargs['scope'], [])

        self.assertEqual(http_response.url, redirect_url)

        callback_redirects = (
            CallbackRedirect.objects.filter(session_key=request.session.session_key)
        )
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, redirect_sub_url)
        self.assertEqual(callback_redirect.session_key, request.session.session_key)
        self.assertEqual(callback_redirect.state, state)

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    @patch('esi.views.app_settings.ESI_OAUTH_LOGIN_URL', ESI_OAUTH_LOGIN_URL)
    @patch('esi.views.OAuth2Session', autospec=True)
    def test_redirect_to_url_w_single_scope(self, mock_OAuth2Session):
        state = 'my_awesome_state'
        mock_OAuth2Session.return_value.authorization_url.return_value = \
            (redirect_url, state)

        request = self.factory.get(redirect_url)
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = sso_redirect(request, scopes='abc')
        self.assertTrue(mock_OAuth2Session.called)
        args, kwargs = mock_OAuth2Session.call_args
        self.assertEqual(kwargs['scope'], ['abc'])

        self.assertEqual(http_response.url, redirect_url)

        callback_redirects = (
            CallbackRedirect.objects.filter(session_key=request.session.session_key)
        )
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, redirect_sub_url)
        self.assertEqual(callback_redirect.session_key, request.session.session_key)
        self.assertEqual(callback_redirect.state, state)

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    @patch('esi.views.app_settings.ESI_OAUTH_LOGIN_URL', ESI_OAUTH_LOGIN_URL)
    @patch('esi.views.OAuth2Session', autospec=True)
    def test_redirect_to_url_w_multiple_scopes(self, mock_OAuth2Session):
        state = 'my_awesome_state'
        mock_OAuth2Session.return_value.authorization_url.return_value = \
            (redirect_url, state)

        request = self.factory.get(redirect_url)
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = sso_redirect(request, scopes=['abc', 'def'])
        self.assertTrue(mock_OAuth2Session.called)
        args, kwargs = mock_OAuth2Session.call_args
        self.assertEqual(kwargs['scope'], ['abc', 'def'])

        self.assertEqual(http_response.url, redirect_url)

        callback_redirects = CallbackRedirect.objects\
            .filter(session_key=request.session.session_key)
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, redirect_sub_url)
        self.assertEqual(callback_redirect.session_key, request.session.session_key)
        self.assertEqual(callback_redirect.state, state)

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    @patch('esi.views.app_settings.ESI_OAUTH_LOGIN_URL', ESI_OAUTH_LOGIN_URL)
    @patch('esi.views.reverse', autospec=True)
    @patch('esi.views.OAuth2Session', autospec=True)
    def test_redirect_to_view_no_scopes(self, mock_OAuth2Session, mock_reverse):
        state = 'my_awesome_state'
        mock_OAuth2Session.return_value.authorization_url.return_value = \
            (redirect_url, state)
        my_view_url = '/my_view/'
        mock_reverse.return_value = my_view_url

        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = sso_redirect(request, return_to='my_view')

        self.assertTrue(mock_OAuth2Session.called)
        args, kwargs = mock_OAuth2Session.call_args
        self.assertEqual(kwargs['scope'], [])

        self.assertEqual(http_response.url, redirect_url)

        callback_redirects = (
            CallbackRedirect.objects.filter(session_key=request.session.session_key)
        )
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, my_view_url)
        self.assertEqual(callback_redirect.session_key, request.session.session_key)
        self.assertEqual(callback_redirect.state, state)

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    @patch('esi.views.reverse')
    def test_sso_redirect_return_to(self, mock_reverse):
        mock_reverse.return_value = '/callback3/'

        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        sso_redirect(request, return_to='callback3')

        callback_redirects = (
            CallbackRedirect.objects.filter(session_key=request.session.session_key)
        )
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, '/callback3/')
        self.assertEqual(callback_redirect.session_key, request.session.session_key)

    @patch('esi.views.app_settings.ESI_SSO_CLIENT_ID', ESI_SSO_CLIENT_ID)
    @patch('esi.views.app_settings.ESI_SSO_CALLBACK_URL', ESI_SSO_CALLBACK_URL)
    def test_sso_redirect_start_session(self):
        request = self.factory.get('https://www.example.com/callback2/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)

        sso_redirect(request)

        callback_redirects = (
            CallbackRedirect.objects.filter(session_key=request.session.session_key)
        )
        self.assertEqual(len(callback_redirects), 1)
        callback_redirect = callback_redirects.first()
        self.assertEqual(callback_redirect.url, '/callback2/')
        self.assertEqual(callback_redirect.session_key, request.session.session_key)


class TesReceiveCallbackView(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )

        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', 'xyz', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    @patch('esi.managers.TokenManager.create_from_request')
    def test_normal(self, mock_create_from_request):
        mock_create_from_request.return_value = self.token

        request = self.factory.get('https://www.example.com?code=abc&state=123')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        callback_redirect = CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            state='123',
            url=redirect_url
        )
        http_response = receive_callback(request)
        callback_redirect.refresh_from_db()
        self.assertIsInstance(http_response, HttpResponseRedirect)
        self.assertEqual(http_response.url, redirect_url)
        self.assertEqual(callback_redirect.token, self.token)

    def test_missing_code(self):
        request = self.factory.get('https://www.example.com?state=123')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = receive_callback(request)
        self.assertIsInstance(http_response, HttpResponseBadRequest)

    def test_missing_state(self):
        request = self.factory.get('https://www.example.com?code=abc')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        http_response = receive_callback(request)
        self.assertIsInstance(http_response, HttpResponseBadRequest)

    def test_missing_callback(self):
        request = self.factory.get('https://www.example.com?code=abc&state=123')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()

        with self.assertRaises(Http404):
            receive_callback(request)


class TestSelectTokenView(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            'Bruce Wayne',
            'abc@example.com',
            'password'
        )
        self.token = _store_as_Token(
            _generate_token(
                character_id=99,
                character_name=self.user.username,
                scopes=['abc', 'xyz', '123']
            ),
            self.user
        )
        self.factory = RequestFactory()
        CallbackRedirect.objects.all().delete()

    def test_should_render_select_page_on_first_call(self):
        # given
        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()
        # when
        response = select_token(request, scopes='abc')
        # then
        self.assertIn("ESI Token Selection", response.content.decode("utf-8"))

    def test_should_render_select_page_on_second_call(self):
        # given
        request = self.factory.get('https://www.example.com/my_view/')
        request.user = self.user
        middleware = SessionMiddleware(HttpResponse)
        middleware.process_request(request)
        request.session.save()
        CallbackRedirect.objects.create(
            session_key=request.session.session_key,
            state='state123',
            token=self.token
        )
        # when
        response = select_token(request, scopes='abc', new=True)
        # then
        self.assertIn("ESI Token Selection", response.content.decode("utf-8"))
